-- script to add customizable license plates to vehicles

LicensePlates = {}
LicensePlates.CONFIGFILENAME = "xml/defaultVehiclesConfig.xml"
LicensePlates.DIRECTORY = g_currentModDirectory
LicensePlates.I3D = "assets/licensePlate.i3d"

-- load events from external files
source(Utils.getFilename("src/events/LicensePlatesVisibilityEvent.lua", LicensePlates.DIRECTORY))
source(Utils.getFilename("src/events/LicensePlatesTextEvent.lua", LicensePlates.DIRECTORY))
source(Utils.getFilename("src/events/LicensePlatesSymbolColorEvent.lua", LicensePlates.DIRECTORY))

-- load and set up gui
source(Utils.getFilename("src/gui/LicensePlatesGui.lua", LicensePlates.DIRECTORY))
g_gui:loadGui(LicensePlates.DIRECTORY.."src/gui/LicensePlatesGui.xml", "LicensePlatesGui", LicensePlatesGui:new())

-- can be added to all vehicles
function LicensePlates.prerequisitesPresent(specializations)
	return true
end

-- adds necessary event listeners for loading, saving, input and multiplayer
function LicensePlates.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", LicensePlates)
	SpecializationUtil.registerEventListener(vehicleType, "saveToXMLFile", LicensePlates)
	SpecializationUtil.registerEventListener(vehicleType, "onReadStream", LicensePlates)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", LicensePlates)
	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", LicensePlates)
end

-- override vehicle selection
function LicensePlates.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeSelected", LicensePlates.getCanBeSelected)
end

-- allow vehicles with license plates to be selected
function LicensePlates:getCanBeSelected(superFunc)
	local spec = self.spec_licensePlates

	return superFunc(self) or spec.licensePlates
end

-- add the input action to the menu
function LicensePlates:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	local spec = self.spec_licensePlates

	if not g_client or not spec.licensePlates then
		return
	end

	self:clearActionEventsTable(spec.actionEvents)
	if isActiveForInput then
		local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.LICENSE_PLATE_CHANGE, self, LicensePlates.changeActionCallback, false, true, false, true, nil)
		g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_LOW)
	end
end

-- load each vehicle
function LicensePlates:onPostLoad(savegame)
	local spec = self.spec_licensePlates

	local filename = LicensePlates.getFilename(self)
	local config = LicensePlates.CONFIG[filename]

	-- mod support
	local key = "vehicle.licensePlates"
	if hasXMLProperty(self.xmlFile, key) then
		config = LicensePlates.parseVehicle(self.xmlFile, key)
	end

	if not config then
		return -- vehicle has no license plates defined
	end

	if #config.licensePlates == 0 and not config.parentFilename then
		g_logManager:error("Vehicle has no own license plates and no parent!")
		return
	end

	local allLicensePlates = {}
	local currentConfig = config

	-- iterative lookup for parent(s)
	while currentConfig do
		for _, licensePlate in ipairs(currentConfig.licensePlates) do
			table.insert(allLicensePlates, licensePlate)
		end

		-- a misconfiguration can cause an endless loop
		currentConfig = LicensePlates.CONFIG[currentConfig.parentFilename]
	end

	if #allLicensePlates == 0 then
		g_logManager:error("Parent vehicle not found or parent vehicle has no license plates!")
		return
	end

	-- now vehicle has at least one license plate
	spec.licensePlates = {}

	-- set up and link all objects from i3d
	for _, lp in ipairs(allLicensePlates) do
		local i3d = g_i3DManager:loadSharedI3DFile(LicensePlates.I3D, LicensePlates.DIRECTORY, false, false)
		local index = lp.isSmall and "0|0|1" or "0|0|0"
		local plate = I3DUtil.indexToObject(i3d, index, self.i3dMappings)

		local linkNode = I3DUtil.indexToObject(self.components, lp.linkNode, self.i3dMappings)
		if lp.linkNode and not linkNode then
			g_logManager:error("Could not find link node: "..lp.linkNode)
			return
		end

		local hiddenNode = I3DUtil.indexToObject(self.components, lp.hiddenNode, self.i3dMappings)
		if lp.hiddenNode and not hiddenNode then
			g_logManager:error("Could not find node to hide: "..lp.hiddenNode)
		end

		setTranslation(plate, lp.translation[1], lp.translation[2], lp.translation[3])
		setRotation(plate, math.rad(lp.rotation[1]), math.rad(lp.rotation[2]), math.rad(lp.rotation[3]))
		setVisibility(plate, false)
			
		link(linkNode or self.rootNode, plate)

		-- make washable
		Washable.addAllSubWashableNodes(self, plate)

		-- wrapper to add information about size and hidden node
		local wrapper = {
			isSmall = lp.isSmall,
			i3d = plate,
			hiddenNode = hiddenNode -- can be nil
		}

		table.insert(spec.licensePlates, wrapper)

		-- i3d can not be used again because index is strange
		delete(i3d)
	end

	if savegame then
		-- load savegame specific information like visibility and text
		local isVisible = getXMLBool(savegame.xmlFile, savegame.key..".licensePlates#isVisible")
		local text = getXMLString(savegame.xmlFile, savegame.key..".licensePlates#text")
		local symbolColor = getXMLString(savegame.xmlFile, savegame.key..".licensePlates#symbolColor")

		LicensePlates.setData(self, isVisible, text, symbolColor)
	else
		-- set default values if not a savegame, nil is handled by subfunctions
		LicensePlates.setData(self, false, config.defaultText, config.defaultSymbolColor)
	end
end

-- save each vehicle
function LicensePlates:saveToXMLFile(xmlFile, key)
	local spec = self.spec_licensePlates

	if not spec.licensePlates then
		return -- vehicle has no license plates
	end

	-- only save attribute if different from default value

	if spec.isVisible then
		setXMLBool(xmlFile, key.."#isVisible", true)
	end

	if spec.text ~= "" then
		setXMLString(xmlFile, key.."#text", spec.text)
	end

	if spec.symbolColor ~= "black" then
		setXMLString(xmlFile, key.."#symbolColor", spec.symbolColor)
	end
end

-- read spec attributes
function LicensePlates:onReadStream(streamId, connection)
	local spec = self.spec_licensePlates

	if not spec.licensePlates then
		return -- vehicle has no license plates
	end

	LicensePlates.setData(self, streamReadBool(streamId), streamReadString(streamId), streamReadString(streamId))
end

-- write spec attributes
function LicensePlates:onWriteStream(streamId, connection)
	local spec = self.spec_licensePlates

	if not spec.licensePlates then
		return -- vehicle has no license plates
	end

	streamWriteBool(streamId, spec.isVisible)
	streamWriteString(streamId, spec.text)
	streamWriteString(streamId, spec.symbolColor)
end

-- show dialog with current data when input action is performed
function LicensePlates:changeActionCallback()
	local spec = self.spec_licensePlates

	local dialog = g_gui:showDialog("LicensePlatesGui")

	dialog.target:setCallback(LicensePlates.setData, self)
	dialog.target:setData(spec.isVisible, spec.text, spec.symbolColor)
end

-- sets all attributes together
function LicensePlates:setData(isVisible, text, symbolColor)
	LicensePlates.setVisible(self, isVisible)
	LicensePlates.setText(self, text)
	LicensePlates.setSymbolColor(self, symbolColor)
end

-- set visibility
function LicensePlates:setVisible(isVisible, noEventSend)
	local spec = self.spec_licensePlates
	isVisible = isVisible == true -- ensures boolean and not nil

	if spec.isVisible == isVisible then
		return -- nothing to change
	end

	spec.isVisible = isVisible

	for _, plate in ipairs(spec.licensePlates) do
		setVisibility(plate.i3d, isVisible)
		if plate.hiddenNode then
			setVisibility(plate.hiddenNode, not isVisible)
		end
	end

	LicensePlatesVisibilityEvent.sendEvent(self, isVisible, noEventSend)
end

-- set the text
function LicensePlates:setText(text, noEventSend, forceUpdate)
	local spec = self.spec_licensePlates
	text = text or "" -- ensures always string

	if spec.text == text and not forceUpdate then
		return -- nothing to change
	end

	spec.text = text

	for _, plate in ipairs(spec.licensePlates) do
		LicensePlates.clearVisual(plate.i3d)
		for i = 1, #text do
			LicensePlates.setVisualChar(plate, i, text:sub(i,i), spec.symbolColor)
		end
	end

	LicensePlatesTextEvent.sendEvent(self, text, noEventSend)
end

-- set symbol color
function LicensePlates:setSymbolColor(symbolColor, noEventSend)
	local spec = self.spec_licensePlates
	symbolColor = symbolColor or "black" -- ensures always color

	if spec.symbolColor == symbolColor then
		return -- nothing to change
	end

	spec.symbolColor = symbolColor

	-- set text again to refresh letters
	LicensePlates.setText(self, spec.text, true, true) -- only local and force update

	LicensePlatesSymbolColorEvent.sendEvent(self, symbolColor, noEventSend)
end

-- removes the text from the plates
function LicensePlates.clearVisual(plate)
	for i = 0, getNumOfChildren(plate) - 1 do
		local placeholder = getChildAt(plate, i)

		if getNumOfChildren(placeholder) > 0 then
			local symbol = getChildAt(placeholder, 0)
			unlink(symbol)
			delete(symbol)
		end
	end
end

-- sets a character at a position from 1 to 10 for a license plates
function LicensePlates.setVisualChar(plate, position, character, symbolColor)
	local index = LicensePlates.getIndex(character, symbolColor)

	if not index then
		return -- user entered unsupported character
	end

	-- load and link new symbol
	local i3d = g_i3DManager:loadSharedI3DFile(LicensePlates.I3D, LicensePlates.DIRECTORY, false, false)
	local symbol = I3DUtil.indexToObject(i3d, index)
	local placeholder = getChildAt(plate.i3d, position - 1)

	-- symbols are translated in i3d
	setTranslation(symbol, 0, 0, 0)

	link(placeholder, symbol)

	delete(i3d)
end

-- gets the index for a character in the i3d
function LicensePlates.getIndex(c, symbolColor)
	local colorPrefix = "0|2|" -- black by default

	if symbolColor == "green" then
		colorPrefix = "0|3|"
	elseif symbolColor == "red" then
		colorPrefix = "0|4|"
	end

	if "a" <= c and c <= "z" then
		-- turns into upper case
		return colorPrefix.."0|" ..(c:byte() - 97)
	elseif "A" <= c and c <= "Z" then
		return colorPrefix.."0|" ..(c:byte() - 65)
	elseif "0" <= c and c <= "9" then
		-- stupid order in i3d so put 0 to the end
		return colorPrefix.."1|" ..((c:byte() - 48 + 9) % 10)
	elseif c == "#" then
		return "0|1|0"
	elseif c == "*" then
		return "0|1|1"
	end

	return nil
end

-- create a unique filename for each vehicle
function LicensePlates:getFilename()
	local modName, baseDir = Utils.getModNameAndBaseDirectory(self.configFileName)

	if modName then
		return modName.."/"..string.gsub(self.configFileName, baseDir, "")
	end

	return self.configFileName
end

-- read the config xml
function LicensePlates.readConfig()
	local xmlFile = loadXMLFile("ConfigXML", Utils.getFilename(LicensePlates.CONFIGFILENAME, LicensePlates.DIRECTORY))

	local vehicles = LicensePlates.parseVehicles(xmlFile, "vehicles")

	delete(xmlFile)

	return vehicles
end

-- parse all default vehicles
function LicensePlates.parseVehicles(xmlFile, superKey)
	local vehicles = {}
	local i = 0

	while true do
		local key = superKey..string.format(".vehicle(%d)", i)
		if not hasXMLProperty(xmlFile, key) then
			break
		end

		local vehicle = LicensePlates.parseVehicle(xmlFile, key)

		if not vehicle.modname or g_modIsLoaded[vehicle.modname] then
			-- faster search by using file as key
			vehicles[vehicle.filename] = vehicle
		end
		-- else mod is not active, do not waste memory and ignore vehicle

		i = i + 1
	end

	return vehicles
end

-- parse a single vehicle including the license plates
function LicensePlates.parseVehicle(xmlFile, key)
	return {
		filename = getXMLString(xmlFile, key.."#filename"),
		defaultText = getXMLString(xmlFile, key.."#defaultText"),
		defaultSymbolColor = getXMLString(xmlFile, key.."#defaultSymbolColor"),
		parentFilename = getXMLString(xmlFile, key..".parent#filename"),
		licensePlates = LicensePlates.parseLicensePlates(xmlFile, key)
	}
end

-- parse all license plates for a vehicle
function LicensePlates.parseLicensePlates(xmlFile, superKey)
	local licensePlates = {}
	local i = 0

	while true do
		local key = superKey..string.format(".licensePlate(%d)", i)
		if not hasXMLProperty(xmlFile, key) then
			break
		end

		table.insert(licensePlates, LicensePlates.parseLicensePlate(xmlFile, key))

		i = i + 1
	end

	return licensePlates

end

-- parse a single license plate
function LicensePlates.parseLicensePlate(xmlFile, key)
	return {
		isSmall = getXMLBool(xmlFile, key.."#isSmall"),
		linkNode = getXMLString(xmlFile, key.."#linkNode"),
		rotation = StringUtil.getVectorNFromString(getXMLString(xmlFile, key.."#rotation") or "0 0 0", 3),
		translation = StringUtil.getVectorNFromString(getXMLString(xmlFile, key.."#translation") or "0 0 0", 3),
		hiddenNode = getXMLString(xmlFile, key.."#hiddenNode"),
	}
end

-- add license plates to all vehicles
function LicensePlates.addSpec()
	local specName = "licensePlates"
	for _, typeDef in pairs(g_vehicleTypeManager.vehicleTypes) do
		typeDef.specializationsByName[specName] = LicensePlates
		table.insert(typeDef.specializationNames, specName)
		table.insert(typeDef.specializations, LicensePlates)
	end
end

-- initialize mod
if g_currentModName == "FS19_LicensePlates" then
	LicensePlates.CONFIG = LicensePlates.readConfig()
	VehicleTypeManager.finalizeVehicleTypes = Utils.prependedFunction(VehicleTypeManager.finalizeVehicleTypes, LicensePlates.addSpec)
else
	LicensePlates = nil
	g_logManager:error("Mod 'LicensePlates' is an invalid version and was not loaded!'")
	g_logManager:error("Please download this mod only from modhub!'")
end
